@@ -5,7 +5,7 @@ import socket
 import time
 from contextlib import suppress
 from dataclasses import astuple, dataclass
-from typing import Any, Callable, Coroutine, List, Optional
+from typing import Any, Callable, Coroutine, Dict, Iterable, List, Optional, Type
 
 import async_timeout
 from google.protobuf import message
@@ -49,6 +49,10 @@ _LOGGER = logging.getLogger(__name__)
 
 BUFFER_SIZE = 1024 * 1024  # Set buffer limit to 1MB
 
+INTERNAL_MESSAGE_TYPES = {GetTimeRequest, PingRequest, DisconnectRequest}
+
+PROTO_TO_MESSAGE_TYPE = {v: k for k, v in MESSAGE_TYPE_TO_PROTO.items()}
+
 
 @dataclass
 class ConnectionParams:
@@ -96,7 +100,7 @@ class APIConnection:
         self._connect_complete = False
 
         # Message handlers currently subscribed to incoming messages
-        self._message_handlers: List[Callable[[message.Message], None]] = []
+        self._message_handlers: Dict[Any, List[Callable[[message.Message], None]]] = {}
         # The friendly name to show for this connection in the logs
         self.log_name = params.address
 
@@ -105,7 +109,7 @@ class APIConnection:
 
         self._ping_stop_event = asyncio.Event()
 
-        self._to_process: asyncio.Queue[message.Message] = asyncio.Queue()
+        self._to_process: asyncio.Queue[Optional[Packet]] = asyncio.Queue()
 
         self._process_task: Optional[asyncio.Task[None]] = None
 
@@ -120,6 +124,9 @@ class APIConnection:
 
         async def _do_cleanup() -> None:
             async with self._connect_lock:
+                # Tell the process loop to stop
+                self._to_process.put_nowait(None)
+
                 if self._frame_helper is not None:
                     await self._frame_helper.close()
                     self._frame_helper = None
@@ -379,19 +386,19 @@ class APIConnection:
         if not self._is_socket_open:
             raise APIConnectionError("Connection isn't established yet")
 
-        for message_type, klass in MESSAGE_TYPE_TO_PROTO.items():
-            if isinstance(msg, klass):
-                break
-        else:
+        message_type = PROTO_TO_MESSAGE_TYPE.get(type(msg))
+        if not message_type:
             raise ValueError(f"Message type id not found for type {type(msg)}")
-
         encoded = msg.SerializeToString()
         _LOGGER.debug("%s: Sending %s: %s", self._params.address, type(msg), str(msg))
 
+        frame_helper = self._frame_helper
+        assert frame_helper is not None
+        if not frame_helper.ready:
+            await frame_helper.wait_for_ready()
+
         try:
-            assert self._frame_helper is not None
-            # pylint: disable=undefined-loop-variable
-            await self._frame_helper.write_packet(
+            frame_helper.write_packet(
                 Packet(
                     type=message_type,
                     data=encoded,
@@ -404,32 +411,47 @@ class APIConnection:
             raise
 
     def add_message_callback(
-        self, on_message: Callable[[Any], None]
+        self, on_message: Callable[[Any], None], msg_types: Iterable[Type[Any]]
     ) -> Callable[[], None]:
         """Add a message callback."""
-        self._message_handlers.append(on_message)
+        for msg_type in msg_types:
+            self._message_handlers.setdefault(msg_type, []).append(on_message)
 
         def unsub() -> None:
-            self._message_handlers.remove(on_message)
+            for msg_type in msg_types:
+                self._message_handlers[msg_type].remove(on_message)
 
         return unsub
 
-    def remove_message_callback(self, on_message: Callable[[Any], None]) -> None:
+    def remove_message_callback(
+        self, on_message: Callable[[Any], None], msg_types: Iterable[Type[Any]]
+    ) -> None:
         """Remove a message callback."""
-        self._message_handlers.remove(on_message)
+        for msg_type in msg_types:
+            self._message_handlers[msg_type].remove(on_message)
 
     async def send_message_callback_response(
-        self, send_msg: message.Message, on_message: Callable[[Any], None]
+        self,
+        send_msg: message.Message,
+        on_message: Callable[[Any], None],
+        msg_types: Iterable[Type[Any]],
     ) -> None:
         """Send a message to the remote and register the given message handler."""
-        self._message_handlers.append(on_message)
-        await self.send_message(send_msg)
+        for msg_type in msg_types:
+            self._message_handlers.setdefault(msg_type, []).append(on_message)
+        try:
+            await self.send_message(send_msg)
+        except (asyncio.CancelledError, Exception):
+            for msg_type in msg_types:
+                self._message_handlers[msg_type].remove(on_message)
+            raise
 
     async def send_message_await_response_complex(
         self,
         send_msg: message.Message,
         do_append: Callable[[message.Message], bool],
         do_stop: Callable[[message.Message], bool],
+        msg_types: Iterable[Type[Any]],
         timeout: float = 10.0,
     ) -> List[message.Message]:
         """Send a message to the remote and build up a list response.
@@ -460,11 +482,15 @@ class APIConnection:
                     new_exc.__cause__ = exc
                 fut.set_exception(new_exc)
 
-        self._message_handlers.append(on_message)
+        for msg_type in msg_types:
+            self._message_handlers.setdefault(msg_type, []).append(on_message)
         self._read_exception_handlers.append(on_read_exception)
-        await self.send_message(send_msg)
+        # We must not await without a finally or
+        # the message could fail to be removed if the
+        # the await is cancelled
 
         try:
+            await self.send_message(send_msg)
             async with async_timeout.timeout(timeout):
                 await fut
         except asyncio.TimeoutError as err:
@@ -472,8 +498,9 @@ class APIConnection:
                 f"Timeout waiting for response for {type(send_msg)} after {timeout}s"
             ) from err
         finally:
-            with suppress(ValueError):
-                self._message_handlers.remove(on_message)
+            for msg_type in msg_types:
+                with suppress(ValueError):
+                    self._message_handlers[msg_type].remove(on_message)
             with suppress(ValueError):
                 self._read_exception_handlers.remove(on_read_exception)
 
@@ -482,11 +509,12 @@ class APIConnection:
     async def send_message_await_response(
         self, send_msg: message.Message, response_type: Any, timeout: float = 10.0
     ) -> Any:
-        def is_response(msg: message.Message) -> bool:
-            return isinstance(msg, response_type)
-
         res = await self.send_message_await_response_complex(
-            send_msg, is_response, is_response, timeout=timeout
+            send_msg,
+            lambda msg: True,  # we will only get responses of `response_type`
+            lambda msg: True,  # we will only get responses of `response_type`
+            (response_type,),
+            timeout=timeout,
         )
         if len(res) != 1:
             raise APIConnectionError(f"Expected one result, got {len(res)}")
@@ -507,71 +535,85 @@ class APIConnection:
             handler(err)
         await self._cleanup()
 
-    async def _read_once(self) -> None:
-        assert self._frame_helper is not None
-        pkt = await self._frame_helper.read_packet()
-
-        msg_type = pkt.type
-        raw_msg = pkt.data
-        if msg_type not in MESSAGE_TYPE_TO_PROTO:
-            _LOGGER.debug("%s: Skipping message type %s", self.log_name, msg_type)
-            return
-
-        msg = MESSAGE_TYPE_TO_PROTO[msg_type]()
-        try:
-            msg.ParseFromString(raw_msg)
-        except Exception as e:
-            raise ProtocolAPIError(f"Invalid protobuf message: {e}") from e
-        _LOGGER.debug("%s: Got message of type %s: %s", self.log_name, type(msg), msg)
-        self._to_process.put_nowait(msg)
-
     async def _process_loop(self) -> None:
+        to_process = self._to_process
         while True:
-            if not self._is_socket_open:
-                # Socket closed but task isn't cancelled yet
-                break
-
             try:
-                msg = await self._to_process.get()
+                pkt = await to_process.get()
             except RuntimeError:
                 break
 
-            for handler in self._message_handlers[:]:
-                handler(msg)
-            await self._handle_internal_messages(msg)
-
-    async def _read_loop(self) -> None:
-        while True:
-            if not self._is_socket_open:
+            if pkt is None:
                 # Socket closed but task isn't cancelled yet
                 break
-            try:
-                await self._read_once()
-            except SocketClosedAPIError as err:
-                # don't log with info, if closed the site that closed the connection should log
+
+            msg_type_proto = pkt.type
+            if msg_type_proto not in MESSAGE_TYPE_TO_PROTO:
                 _LOGGER.debug(
-                    "%s: Socket closed, stopping read loop",
-                    self.log_name,
+                    "%s: Skipping message type %s", self.log_name, msg_type_proto
                 )
-                await self._report_fatal_error(err)
-                break
-            except APIConnectionError as err:
-                _LOGGER.info(
-                    "%s: Error while reading incoming messages: %s",
-                    self.log_name,
-                    err,
+                continue
+
+            msg = MESSAGE_TYPE_TO_PROTO[msg_type_proto]()
+            try:
+                msg.ParseFromString(pkt.data)
+            except Exception as e:
+                await self._report_fatal_error(
+                    ProtocolAPIError(f"Invalid protobuf message: {e}")
                 )
-                await self._report_fatal_error(err)
-                break
-            except Exception as err:  # pylint: disable=broad-except
-                _LOGGER.warning(
-                    "%s: Unexpected error while reading incoming messages: %s",
-                    self.log_name,
-                    err,
-                    exc_info=True,
-                )
-                await self._report_fatal_error(err)
-                break
+                raise
+
+            msg_type = type(msg)
+
+            _LOGGER.debug(
+                "%s: Got message of type %s: %s", self.log_name, msg_type, msg
+            )
+
+            for handler in self._message_handlers.get(msg_type, [])[:]:
+                handler(msg)
+
+            # Pre-check the message type to avoid awaiting
+            # since most messages are not internal messages
+            if msg_type in INTERNAL_MESSAGE_TYPES:
+                await self._handle_internal_messages(msg)
+
+    async def _read_loop(self) -> None:
+        frame_helper = self._frame_helper
+        assert frame_helper is not None
+        await frame_helper.wait_for_ready()
+        to_process = self._to_process
+        try:
+            # Once its ready, we hold the lock for the duration of the
+            # connection so we don't have to keep locking/unlocking
+            async with frame_helper.read_lock:
+                while True:
+                    to_process.put_nowait(await frame_helper.read_packet_with_lock())
+        except SocketClosedAPIError as err:
+            # don't log with info, if closed the site that closed the connection should log
+            if not self._is_socket_open:
+                # If we expected the socket to be closed, don't log
+                # the error.
+                return
+            _LOGGER.debug(
+                "%s: Socket closed, stopping read loop",
+                self.log_name,
+            )
+            await self._report_fatal_error(err)
+        except APIConnectionError as err:
+            _LOGGER.info(
+                "%s: Error while reading incoming messages: %s",
+                self.log_name,
+                err,
+            )
+            await self._report_fatal_error(err)
+        except Exception as err:  # pylint: disable=broad-except
+            _LOGGER.warning(
+                "%s: Unexpected error while reading incoming messages: %s",
+                self.log_name,
+                err,
+                exc_info=True,
+            )
+            await self._report_fatal_error(err)
 
     async def _handle_internal_messages(self, msg: Any) -> None:
         if isinstance(msg, DisconnectRequest):
