@@ -1,5 +1,6 @@
 import enum
 from dataclasses import asdict, dataclass, field, fields
+from functools import cache
 from typing import (
     TYPE_CHECKING,
     Any,
@@ -51,10 +52,15 @@ class APIIntEnum(enum.IntEnum):
         return ret
 
 
+# Fields do not change so we can cache the result
+# of calling fields() on the dataclass
+cached_fields = cache(fields)
+
+
 @dataclass(frozen=True)
 class APIModelBase:
     def __post_init__(self) -> None:
-        for field_ in fields(type(self)):
+        for field_ in cached_fields(type(self)):  # type: ignore[arg-type]
             convert = field_.metadata.get("converter")
             if convert is None:
                 continue
@@ -71,15 +77,14 @@ class APIModelBase:
     ) -> _V:
         init_args = {
             f.name: data[f.name]
-            for f in fields(cls)
+            for f in cached_fields(cls)  # type: ignore[arg-type]
             if f.name in data or (not ignore_missing)
         }
         return cls(**init_args)
 
     @classmethod
     def from_pb(cls: Type[_V], data: Any) -> _V:
-        init_args = {f.name: getattr(data, f.name) for f in fields(cls)}
-        return cls(**init_args)
+        return cls(**{f.name: getattr(data, f.name) for f in cached_fields(cls)})  # type: ignore[arg-type]
 
 
 def converter_field(*, converter: Callable[[Any], _V], **kwargs: Any) -> _V:
@@ -758,11 +763,6 @@ class UserService(APIModelBase):
 
 
 # ==================== BLUETOOTH ====================
-def _long_uuid(uuid: str) -> str:
-    """Convert a UUID to a long UUID."""
-    return (
-        f"0000{uuid[2:].lower()}-0000-1000-8000-00805f9b34fb" if len(uuid) < 8 else uuid
-    ).lower()
 
 
 def _join_split_uuid(value: List[int]) -> str:
@@ -770,8 +770,17 @@ def _join_split_uuid(value: List[int]) -> str:
     return str(UUID(int=((value[0] << 64) | value[1])))
 
 
-def _convert_bluetooth_le_service_uuids(value: List[str]) -> List[str]:
-    return [_long_uuid(v) for v in value]
+# value is likely a google.protobuf.pyext._message.RepeatedScalarContainer
+def _convert_bluetooth_le_service_uuids(value: Iterable[str]) -> List[str]:
+    if not value:
+        # empty list, don't convert
+        return []
+
+    # Long UUID inlined to avoid call stack inside the list comprehension
+    return [
+        f"0000{v[2:].lower()}-0000-1000-8000-00805f9b34fb" if len(v) < 8 else v.lower()
+        for v in value
+    ]
 
 
 def _convert_bluetooth_le_service_data(
@@ -780,7 +789,17 @@ def _convert_bluetooth_le_service_data(
     if isinstance(value, dict):
         return value
 
-    return {_long_uuid(v.uuid): bytes(v.data if v.data else v.legacy_data) for v in value}  # type: ignore
+    if not value:
+        return {}
+
+    # Long UUID inlined to avoid call stack inside the dict comprehension
+    return {
+        f"0000{v.uuid[2:].lower()}-0000-1000-8000-00805f9b34fb"  # type: ignore[union-attr]
+        if len(v.uuid) < 8  # type: ignore[union-attr]
+        # v.data if v.data else v.legacy_data is backwards compatible with ESPHome devices before 2022.10.0
+        else v.uuid.lower(): bytes(v.data if v.data else v.legacy_data)  # type: ignore[union-attr]
+        for v in value
+    }
 
 
 def _convert_bluetooth_le_manufacturer_data(
@@ -788,7 +807,11 @@ def _convert_bluetooth_le_manufacturer_data(
 ) -> Dict[int, bytes]:
     if isinstance(value, dict):
         return value
-    # v.data if v.data else v.legacy_data is backwards compatable with ESPHome devices before 2022.10.0
+
+    if not value:
+        return {}
+
+    # v.data if v.data else v.legacy_data is backwards compatible with ESPHome devices before 2022.10.0
     return {int(v.uuid, 16): bytes(v.data if v.data else v.legacy_data) for v in value}  # type: ignore
 
 
@@ -797,6 +820,7 @@ class BluetoothLEAdvertisement(APIModelBase):
     address: int = 0
     name: str = ""
     rssi: int = 0
+    address_type: int = 0
 
     service_uuids: List[str] = converter_field(
         default_factory=list, converter=_convert_bluetooth_le_service_uuids
@@ -913,6 +937,8 @@ class BluetoothDeviceRequestType(APIIntEnum):
     DISCONNECT = 1
     PAIR = 2
     UNPAIR = 3
+    CONNECT_V3_WITH_CACHE = 4
+    CONNECT_V3_WITHOUT_CACHE = 5
 
 
 class LogLevel(APIIntEnum):
